#!/usr/bin/env python
#
#  This file is a part of pyscmclient, a Python based command line tool for an 
#  SCM-Manager server (http://www.scm-manager.org) 
#
#  Copyright (c) 2013 Jori Liesenborgs
#  Contact: jori.liesenborgs@gmail.com
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#  IN THE SOFTWARE.

import pprint
import logging
import argparse
import sys
import time
import getpass
from scmconnection import SCMException, SCMConnection

def dateString(milliseconds):
    if not milliseconds:
        return ""
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(milliseconds/1000.0))

def addGroupMembers(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Add members to the specified group")
    confParser.add_argument('groupname', help="Name of the group to add members to")
    confParser.add_argument('membernames', nargs=argparse.REMAINDER, help="Names of the members to add")
    args = confParser.parse_args(args)

    scm.addMembers(args.groupname, args.membernames)
    
    getGroup("get-group", scm, [ args.groupname ])

def delGroupMembers(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Delete members from the specified group")
    confParser.add_argument('groupname', help="Name of the group to delete members from")
    confParser.add_argument('membernames', nargs=argparse.REMAINDER, help="Names of the members to delete")
    args = confParser.parse_args(args)

    scm.delMembers(args.groupname, args.membernames)
    
    getGroup("get-group", scm, [ args.groupname ])

def listGroups(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "List the existing groups")
    args = confParser.parse_args(args)

    results = scm.getGroups()

    for r in results:
        sys.stdout.write(r["name"] + ":")
        for i in range(len(r["members"])):
            m = r["members"][i]
            m = m.strip()
            if i != 0:
                s = ",%s" % m
            else:
                s = "%s" % m
            sys.stdout.write(s)
        sys.stdout.write("\n")

def getGroup(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Display detailed information about a group")
    confParser.add_argument("groupname", help="Name of the group")
    args = confParser.parse_args(args)

    curGroup = scm.getGroup(args.groupname)

    print "Name:           %s" % curGroup["name"]
    print "Description:    %s" % curGroup["description"]
    print "Creation-Date:  %s" % dateString(curGroup["creationDate"])
    print "Last-Modified:  %s" % dateString(curGroup["lastModified"])
    print "Members:"
    for m in curGroup["members"]:
        print "  %s" % m

    print

def listUsers(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "List the currently defined user (admin only)")
    args = confParser.parse_args(args)

    results = scm.getUsers()
    for r in results:
        print "%s,%s,%s" % (r["name"],r["displayName"],str(r["admin"]))

def searchUser(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "List the user names which match the specified search string")
    confParser.add_argument("searchstring", help="Search string for user name")
    args = confParser.parse_args(args)

    userNames = scm.searchUser(args.searchstring)
    
    if not userNames:
        print "No users found"
    else:
        print "Matching users (not all matches may be shown):"
        # Find the maximum length first
        maxLen = 0
        for u in userNames:
            if len(u[0]) > maxLen:
                maxLen = len(u[0])

        for u in userNames:
            print "   " + u[0] + ": " + (" " * (maxLen-len(u[0]))) + u[1]

def getUser(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Get the information about a user (admin only)")
    confParser.add_argument("username", help="User login name")
    args = confParser.parse_args(args)

    curUser = scm.getUser(args.username)
    
    print "Name:           %s" % curUser["name"]
    print "Display Name:   %s" % curUser["displayName"]
    print "E-Mail:         %s" % curUser["mail"]
    print "Administrator:  %s" % str(curUser["admin"])
    print "Creation-Date:  %s" % dateString(curUser["creationDate"])
    print "Last-Modified:  %s" % dateString(curUser["lastModified"])

def createUser(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Create a new user with specified password")
    confParser.add_argument("username", help="User login name")
    confParser.add_argument("password", help="User login password")
    confParser.add_argument("-a", "--admin", help="Flag indicating if this is an administrator", action="store_true")
    confParser.add_argument("-d", "--displayname", help="Display name", default="New User")
    confParser.add_argument("-m", "--mail", help="E-mail address", default="unknown@invalid.location.net")
    args = confParser.parse_args(args)

    logging.debug("Obtaining current users")
    results = scm.getUsers()
    curUsers = [ r["name"] for r in results ]

    if args.username in curUsers:
        raise SCMException("User '%s' already exists in the system" % args.username)

    scm.createUser(args.username, args.password, args.displayname, args.mail, args.admin)

def listRepositories(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "List the repositories to which you have access")
    args = confParser.parse_args(args)

    results = scm.getRepositories()
    for r in results:
        print "%s,%s,%s" % (r["type"], r["id"], r["name"])

    #pprint.pprint(results)

def getRepository(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Get detailed information about the repository with specified ID")
    confParser.add_argument("id", help="ID of the repository")
    args = confParser.parse_args(args)

    curRepo = scm.getRepository(args.id)
    
    print "ID:             %s" % curRepo["id"]
    print "Name:           %s" % curRepo["name"]
    print "Type:           %s" % curRepo["type"]
    print "E-Mail:         %s" % curRepo["contact"]
    print "Description:    %s" % curRepo["description"]
    print "Public:         %s" % str(curRepo["public"])
    print "Creation-Date:  %s" % dateString(curRepo["creationDate"])
    print "Last-Modified:  %s" % dateString(curRepo["lastModified"])
    print "URL:            %s" % curRepo["url"]

    if curRepo["permissions"] is None:
        print "Permissions:    Not allowed to see"
    else:
        print "Permissions:"
        for p in curRepo["permissions"]:
            print "  %s - %s (Group: %s)" % (p["type"], p["name"], str(p["groupPermission"]))

    print

def createRepository(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Create a repository (as admin or as user if allowed)")
    confParser.add_argument("type", help="Repository type, can be 'hg', 'git' or 'svn'")
    confParser.add_argument("name", help="Name of the repository (may be changed by the extension)")
    confParser.add_argument("-c", "--contact", help="Contact for the repository", default = "")
    confParser.add_argument("-d", "--description", help="Description for the repository", default = "")
    confParser.add_argument("-p", "--public", help="Make this a public repository", action="store_true")
    args = confParser.parse_args(args)

    newRepo = scm.createRepository(args.type, args.name, args.contact, args.description, args.public)

    repoId = newRepo["id"]
    repoName = newRepo["name"]

    print "%s,%s" % (repoId,repoName)

def deleteRepository(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Delete the repository with the specified ID")
    confParser.add_argument("id", help="ID of the repository")
    confParser.add_argument("--reallyreallysure", help="You need to specify this flag to delete the specified repository", action="store_true")

    args = confParser.parse_args(args)

    getRepository("get-repository", scm, [ args.id ])

    if not args.reallyreallysure:
        print "You need to specify --reallyreallysure to actually delete this repository"
    else:
        scm.deleteRepository(args.id)

def setRepositoryPermission(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Add/modify user or group permission settings for the repository with the specified ID")
    confParser.add_argument("id", help="ID of the repository")
    confParser.add_argument("name", help="User/group name for which to set the permission")
    confParser.add_argument("type", help="Can be OWNER, READ or WRITE")
    confParser.add_argument("-g", "--group", help="Set this flag if the name refers to a group instead of a user", action="store_true")
    args = confParser.parse_args(args)

    scm.setRepositoryPermission(args.id, args.group, args.name, args.type)

    getRepository("get-repository", scm, [ args.id ])

def removeRepositoryPermission(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Remove user or group permissions for the repository with specified ID")
    confParser.add_argument("id", help="ID of the repository")
    confParser.add_argument("name", help="User/group name for which to set the permission")
    confParser.add_argument("-g", "--group", help="Set this flag if the name refers to a group instead of a user", action="store_true")
    args = confParser.parse_args(args)

    scm.removeRepositoryPermission(args.id, args.group, args.name)

    getRepository("get-repository", scm, [ args.id ])

def verifyVariousThings(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "Verify that users and groups mentioned in repositories are valid (doesn't make changes)")
    args = confParser.parse_args(args)

    usersInfo = scm.getUsers()
    groupsInfo = scm.getGroups()
    repoInfo = scm.getRepositories()

    users = set()
    for u in usersInfo:
        users.add(u["name"])
    
    groups = set()
    for g in groupsInfo:
        groups.add(g["name"])

    ok = True

    for r in repoInfo:
        repoId = r["id"]
        repoName = r["name"]
        permissions = r["permissions"]
        for p in permissions:
            name = p["name"]
            if p["groupPermission"]:
                if not name in groups:
                    print "Warning: In '%s' (%s): group '%s' mentioned but doesn't exist" % (repoName, repoId, name)
                    ok = False
            else:
                if not name in users:
                    print "Warning: In '%s' (%s): user '%s' mentioned but doesn't exist" % (repoName, repoId, name)
                    ok = False
                
    if ok:
        print "Verification ok"

def listUserRepositories(cmdName, scm, args):
    confParser = argparse.ArgumentParser(prog = cmdName, add_help = True, description = "List the repositories to which a specific user has access (only directly, not by group membership)")
    confParser.add_argument("username", help="User login name")
    args = confParser.parse_args(args)

    try:
        userInfo = scm.getUser(args.username)
    except Exception as e:
        logging.warning("Warning: can't get user info for '%s' (user doesn't exit? not administrator?)" % args.username)
    
    repositories = scm.getRepositories()
    for r in repositories:
        repoId = r["id"]
        repoName = r["name"]
        permissions = r["permissions"]
        for p in permissions:
            if not p["groupPermission"] and p["name"] == args.username:
                print "%s,%s,%s" % (repoId, repoName, p["type"])


def main():
    commands = { 
                 "repo-list": listRepositories, 
                 "repo-get": getRepository,
                 "repo-create": createRepository,
                 "repo-delete": deleteRepository,
                 "repo-perm-set": setRepositoryPermission,
                 "repo-perm-del": removeRepositoryPermission,
                 "user-list": listUsers,
                 "user-get": getUser,
                 "user-create": createUser,
                 "user-search": searchUser,
                 "user-repo-list": listUserRepositories,
                 "group-list": listGroups,
                 "group-get": getGroup,
                 "group-members-add": addGroupMembers,
                 "group-members-del": delGroupMembers,
                 "verify": verifyVariousThings,
               }

    epilog = "Possible commands are:\n"
    sortedCommands = [ i for i in commands ]
    sortedCommands.sort()
    for i in sortedCommands:
        epilog += "  " + i + "\n"
    epilog += "\n"

    confParser = argparse.ArgumentParser(description = "Connect to an scm-manager server and execute a command",
                                         epilog = epilog, add_help = True, formatter_class=argparse.RawDescriptionHelpFormatter)
    confParser.add_argument("-u", "--username", help="Login name for the scm-manager system", default = "anonymous")
    confParser.add_argument("-p", "--password", help="Password needed to log in", default = "")
    confParser.add_argument("-s", "--server", help="URL of the scm-manager server (e.g. http://localhost:8080/scm) ", default = "http://localhost:8080/scm")
    confParser.add_argument("-v", "--verbose", help="Enable more log messages, useful for debugging", action="store_true")
    confParser.add_argument("commandname", help="The name of the command to execute")
    confParser.add_argument('args', nargs=argparse.REMAINDER, help="Command specific arguments")

    args = confParser.parse_args()

    loglevel = logging.INFO
    if args.verbose:
        loglevel = logging.DEBUG

    #logging.basicConfig(level=loglevel, format="%(asctime)-15s | %(levelname)-8s | %(message)s")
    logging.basicConfig(level=loglevel, format="%(message)s")

    if not args.commandname in commands:
        print "Invalid command name, allowed are:"
        for i in sortedCommands:
            print "    " + i
        sys.exit(-1)

    scm = None
    password = args.password

    if not password:
        password = getpass.getpass()

    try:
        scm = SCMConnection(args.server)
        scm.login(args.username, password)
        
        f = commands[args.commandname]
        f(args.commandname, scm, args.args)

    except Exception as e:
        print e
        sys.exit(-1)
    finally:
        if scm:
            try: scm.logout()
            except: pass

    sys.exit(0)

if __name__ == "__main__":
    main()

