#!/usr/bin/env python
#
#  This file is a part of pyscmclient, a Python based command line tool for an 
#  SCM-Manager server (http://www.scm-manager.org) 
#
#  Copyright (c) 2013 Jori Liesenborgs
#  Contact: jori.liesenborgs@gmail.com
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#  IN THE SOFTWARE.

import urllib
import urllib2    
import cookielib            
import json
import pprint
import copy
import logging

class SCMException(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

def slashIt(s):
    if s[-1] != '/':
        return s + "/"
    return s

class SCMConnection(object):

    def __init__(self, urlBase):

        self.cookieJar = cookielib.LWPCookieJar()
        self.opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(self.cookieJar))
        self.headers =  {'User-agent' : 'Mozilla/5.0 (X11; Linux x86_64; rv:15.0) Gecko/20100101 Firefox/15.0.1' }
        self.urlBase = slashIt(urlBase) + "api/rest/"
        self.isAdmin = False
        self.username = None

    def getData(self, subUrl, queryDict = None, responseJSON = True):

        url = self.urlBase + subUrl

        if queryDict is not None:
            url += "?" + urllib.urlencode(queryDict)

        req = urllib2.Request(url, None, self.headers)

        handle = self.opener.open(req)
        result = handle.read()
        handle.close()
        
        if responseJSON: 
            result = json.loads(result)

        return (result, handle.info())

    def deleteData(self, subUrl, responseJSON = True):
        req = urllib2.Request(self.urlBase + subUrl, None, self.headers)
        req.get_method = lambda: 'DELETE'

        handle = self.opener.open(req)
        result = handle.read()
        handle.close()
        
        if responseJSON: 
            result = json.loads(result)

        return (result, handle.info())

    def postDictData(self, d, subUrl, responseJSON = True):
        postdata = urllib.urlencode(d)
        req = urllib2.Request(self.urlBase + subUrl, postdata, self.headers)

        handle = self.opener.open(req)
        result = handle.read()
        handle.close()

        if responseJSON:
            result = json.loads(result)
        return (result, handle.info())

    def postJSONDictData(self, d, subUrl, responseJSON = True, usePut = False):
        postdata = json.dumps(d)
        headers = copy.copy(self.headers)
        headers["Content-type"] = "application/json; charset=UTF-8"
        req = urllib2.Request(self.urlBase + subUrl, postdata, headers)

        if usePut:
            req.get_method = lambda: 'PUT'

        handle = self.opener.open(req)
        result = handle.read()
        handle.close()

        if responseJSON:
            result = json.loads(result)
        return (result, handle.info())

    def login(self, login, password):

        try:
            (result, info) = self.postDictData({"username": login, "password": password}, "authentication/login.json")
        except Exception as e:
            raise SCMException("Couldn't log in (wrong password or login?): " + str(e))

        self.isAdmin = result["user"]["admin"]
        self.username = copy.copy(login)
        logging.debug("Logged in '%s' as '%s' <%s> (admin = %s)" % (login, str(result["user"]["displayName"]), str(result["user"]["mail"]), str(self.isAdmin)))

    def logout(self):

        try:
            self.getData("authentication/logout.json", responseJSON = False)
        except Exception as e:
            raise SCMException("Couldn't log out: " + str(e))

    def getRepositories(self):
        try:
            (result, info) = self.getData("repositories.json")
        except Exception as e:
            raise SCMException("Couldn't get repositories: " + str(e))
        return result

    def getRepository(self, repoID):

        result = self.getRepositories()
        for r in result:
            if r["id"] == repoID:
                return r

        raise SCMException("Repository with ID '%s' is not found" % repoID)

    def createRepository(self, repoType, repoName, contact = "", description = "", public = False):

        if not repoType in [ "hg", "git", "svn" ]:
            raise SCMException("Invalid repository type (must be 'hg', 'git' or 'svn')")

        url = "repositories.json" if self.isAdmin else "plugins/userrepo/create.json"

        repoInfo = {"name": repoName, "type": repoType, "contact": contact, "description": description, "public": public}

        if not self.isAdmin:
            
            # Obtain the old repository id's so we can try to find out what was added
            
            results = self.getRepositories()
            oldRepos = [ r["id"] for r in results ]

        try:
            (result, info) = self.postJSONDictData(repoInfo, url, False)
        except Exception as e:
            raise SCMException("Unable to add repository (not allowed? name exists?): " +  str(e))

        newRepo = None

        results = self.getRepositories()

        if self.isAdmin: # In this case, the new ID is in the info
            try:
                loc = info["Location"]
                idx = loc.rfind("/")
                newId = loc[idx+1:]
            except Exception as e:
                raise SCMException("Unable to get the new repository ID from the response: " + str(e))

            for r in results:
                if r["id"] == newId:
                    newRepo = r
                    break

            if not newRepo:
                raise SCMException("Repository created, but the returned ID was not found")

            if not repoName in newRepo["name"]:
                raise SCMException("User repository created successfully, but can't figure out the name and id: specified name does not seem to appear in the new repository name")

        else:
            # The creation of a user repository doesn't send back an ID, and the name can be modified,
            # so we'll need to compare the new repo list to the old one

            logging.info("Warning: creating a user repository does not directly send back an ID for the created repository, we're _trying_ to figure out what it is")

            newRepos = set()
            for r in results:
                newRepos.add(r["id"])

            try:
                for rid in oldRepos:
                    newRepos.remove(rid)
            except:
                raise SCMException("Repository created successfully, but can't figure out the name and id: found id in old repo list that is not present in new list")

            newIds = [ i for i in newRepos ]

            if len(newIds) != 1:
                raise SCMException("Repository created successfully, but can't figure out the name and id: appears to be more than one new repository")

            for i in results:
                if i["id"] == newIds[0]:
                    newRepo = i
                    break

        return newRepo

    def deleteRepository(self, repoId):

        # Check that it exists
        repo = self.getRepository(repoId)
        
        self.deleteData("repositories/%s.json" % repoId, False)

    def setRepositoryPermission(self, repoId, isGroup, name, permType):

        if not permType in [ "OWNER", "READ", "WRITE" ]:
            raise SCMException("Permission type must be either OWNER, READ or WRITE")

        repo = self.getRepository(repoId)

        if repo["permissions"] is None:
            raise SCMException("Can't get the permission list for this repository")

        # Get the current permissions

        permissions = { }
        for p in repo["permissions"]:
            identifier = (p["name"], p["groupPermission"])
            permissions[identifier] = p["type"]

        if isGroup: # Check that the name is a valid group name
            groupInfo = self.getGroup(name)
        else: # check that the name is a user
            if self.isAdmin:
                userInfo = self.getUser(name) # Admin can get the full user list
            else: 
                # As a normal user, we'll use the search function (like when adding a user in the web interface)
                # But at the moment this returns at most 5 results, and so the best match may not be in here
                userNames = self.searchUser(name)
                found = False
                for u in userNames:
                    if u[0] == name:
                        found = True
                        break

                maxResults = 5 # From tests and the scm-server source code

                if not found:
                    if len(userNames) < maxResults: # Server definitely didn't find more matches
                        raise SCMException("The specified user name '%s' does not appear to be valid" % name)
                    else: # Just print a warning
                        logging.warning("Warning: can't be certain that user '%s' exists since there are many similar matches" % name)

        # Try to prevent that we're disabling ourselves as OWNER if we're a regular user
        if not self.isAdmin and not isGroup and name == self.username and permType != "OWNER":
            raise SCMException("Trying to disable our own OWNER permission, bailing out! (use the web interface if you're sure)")

        # Add/change the permissions

        newId = (name, isGroup)
        permissions[newId] = permType

        newPermissionList = [ ]
        for permId in permissions:
            value = permissions[permId]
            newPermissionList.append({"type": value, "name": permId[0], "groupPermission": permId[1]})

        # Change the permission list and upload

        repo["permissions"] = newPermissionList

        (result, info) = self.postJSONDictData(repo, "repositories/%s.json" % repoId, responseJSON = False, usePut = True)

    def removeRepositoryPermission(self, repoId, isGroup, name):

        repo = self.getRepository(repoId)

        if repo["permissions"] is None:
            raise SCMException("Can't get the permission list for this repository")

        # Get the current permissions

        permissions = { }
        for p in repo["permissions"]:
            identifier = (p["name"], p["groupPermission"])
            permissions[identifier] = p["type"]

        # Try to prevent that we're disabling ourselves as OWNER if we're a regular user
        if not self.isAdmin and not isGroup and name == self.username:
            raise SCMException("Trying to remove our own OWNER permission, bailing out! (use the web interface if you're sure)")

        # Add/change the permissions

        permId = (name, isGroup)
        try:
            del permissions[permId]
        except:
            # Nothing needs to be done, the (user,group) is not present
            return

        newPermissionList = [ ]
        for permId in permissions:
            value = permissions[permId]
            newPermissionList.append({"type": value, "name": permId[0], "groupPermission": permId[1]})

        # Change the permission list and upload

        repo["permissions"] = newPermissionList

        (result, info) = self.postJSONDictData(repo, "repositories/%s.json" % repoId, responseJSON = False, usePut = True)

    def getUsers(self):
        try:
            (result, info) = self.getData("users.json")
        except Exception as e:
            raise SCMException("Couldn't get user list: " + str(e))
        return result

    def getUser(self, username):

        results = self.getUsers()
        curUser = None

        for r in results:
            if r["name"] == username:
                curUser = r
                break

        if curUser is None:
            raise SCMException("User '%s' is not known" % username)

        return curUser

    def createUser(self, username, password, displayname, email, isAdmin = False):

        userInfo = {
                     "name": username,
                     "displayName": displayname,
                     "mail": email,
                     "password": password,
                     "admin": isAdmin,
                     "active": True,
                     "type": "xml" 
                   }

        try:
            (result, info) = self.postJSONDictData(userInfo, "users.json", False)
        except Exception as e:
            raise SCMException("Couldn't create new user (user exists? make sure email address looks ok and a display name is specified): " + str(e))

    def getGroups(self):
        try:
            (result, info) = self.getData("groups.json")
        except Exception as e:
            raise SCMException("Couldn't get group list: " + str(e))
        return result

    def getGroup(self, groupname):

        results = self.getGroups()
        curGroup = None

        for r in results:
            if r["name"] == groupname:
                curGroup = r
                break

        if curGroup is None:
            raise SCMException("Group '%s' is not known" % groupname)

        return curGroup

    def addMembers(self, groupname, memberlist):
        
        targetGroup = self.getGroup(groupname)

        # Check if users exist

        users = self.getUsers()
        userNames = set()
        for u in users:
            userNames.add(u["name"])

        for m in memberlist:
            if not m in userNames:
                raise SCMException("Specified member '%s' is not a valid username" % m)

        # Add the new members to the old members

        oldmembers = targetGroup["members"]
        memberNames = set()
        for m in oldmembers:  # We're not going to check if the old members are still valid
            memberNames.add(m)

        for m in memberlist:
            if m in memberNames:
                print "Warning: user '%s' is already present in group '%s', will not store duplicates" % (m, groupname)
            memberNames.add(m)

        newMembers = [ m for m in memberNames ]
        targetGroup["members"] = sorted(newMembers)

        try:
            (result, info) = self.postJSONDictData(targetGroup, "groups/%s.json" % groupname, responseJSON = False, usePut = True)
        except Exception as e:
            raise SCMException("Unable to add specified members to group '%s': %s" % (groupname, str(e)))

    def delMembers(self, groupname, memberlist):
        
        targetGroup = self.getGroup(groupname)

        oldmembers = targetGroup["members"]
        memberNames = set()
        for m in oldmembers:  # We're not going to check if the old members are still valid
            memberNames.add(m)

        for m in memberlist:
            if m in memberNames:
                memberNames.remove(m)

        newMembers = [ m for m in memberNames ]
        targetGroup["members"] = sorted(newMembers)

        try:
            (result, info) = self.postJSONDictData(targetGroup, "groups/%s.json" % groupname, responseJSON = False, usePut = True)
        except Exception as e:
            raise SCMException("Unable to delete specified members from group '%s': %s" % (groupname, str(e)) )

    def searchUser(self, searchString):

        query = { "query": searchString }

        try:
            (result, info) = self.getData("search/users.json", query)
        except Exception as e:
            raise SCMException("Unable to search for specified string '%s' in user list: %s" % (searchString, str(e)))

        userList = result["results"]
        if not userList:
            return [ ]

        ret = [ (e["value"], e["label"]) for e in userList ]

        return ret

